@@ -11,18 +11,16 @@ extensions = [
     'sphinx.ext.napoleon',
     'sphinx.ext.todo',
     'sphinx.ext.viewcode',
-    'recommonmark',
 ]
 source_suffix = {
     '.rst': 'restructuredtext',
-    '.md': 'markdown',
 }
 master_doc = 'index'
 project = 'cuTWED'
 year = '2020'
 author = 'Garrett Wright'
 copyright = '{0}, {1}'.format(year, author)
-version = release = '0.3.0'
+version = release = '1.1.0'
 
 pygments_style = 'trac'
 templates_path = ['.']
